from django.contrib.auth.models import User, Group
from rest_framework import serializers
from rest_hooks.models import Hook
from .models import Identity, OptOut, OptIn


class UserSerializer(serializers.ModelSerializer):
    class Meta:
        model = User
        fields = ('url', 'username', 'email', 'groups')


class GroupSerializer(serializers.ModelSerializer):
    class Meta:
        model = Group
        fields = ('url', 'name')


class CreateUserSerializer(serializers.Serializer):
    email = serializers.EmailField()


class IdentitySerializer(serializers.ModelSerializer):
    class Meta:
        model = Identity
        read_only_fields = ('created_at', 'updated_at')
        fields = ('id', 'version', 'details',
                  'communicate_through', 'operator',
                  'created_at', 'created_by', 'updated_at', 'updated_by')


class OptOutSerializer(serializers.ModelSerializer):
    class Meta:
        model = OptOut
        fields = ('id', 'optout_type', 'identity', 'address_type', 'address',
                  'request_source', 'requestor_source_id',
                  'reason', 'created_at')
        read_only_fields = ('created_by')


class OptInSerializer(serializers.ModelSerializer):
    class Meta:
        model = OptIn
        fields = ('id', 'identity', 'address_type', 'address',
                  'request_source', 'requestor_source_id', 'created_at')
        read_only_fields = ('created_by')


class HookSerializer(serializers.ModelSerializer):
    class Meta:
        model = Hook
        read_only_fields = ('user',)


class AddressSerializer(serializers.Serializer):
    address = serializers.CharField(max_length=500)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.1 on 2016-09-26 10:53
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('identities', '0004_detailkey'),
    ]

    operations = [
        migrations.CreateModel(
            name='OptIn',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, help_text='UUID of this opt-in request.', primary_key=True, serialize=False)),
                ('address_type', models.CharField(default='', help_text='Address type used to identify the identity.', max_length=50)),
                ('address', models.CharField(default='', help_text='Address used to identify the identity.', max_length=255)),
                ('request_source', models.CharField(help_text='Service that the optin was requested from.', max_length=100)),
                ('requestor_source_id', models.CharField(help_text='ID for the user requesting the optin on the service that it was requested from. Ideally a UUID.', max_length=500, null=True)),
                ('created_at', models.DateTimeField(auto_now_add=True, help_text='Time request was received.')),
                ('created_by', models.ForeignKey(help_text='User creating the OptIn', null=True, on_delete=django.db.models.deletion.CASCADE, related_name='optin_created', to=settings.AUTH_USER_MODEL)),
                ('identity', models.ForeignKey(help_text='UUID for the identity opting in.', null=True, on_delete=django.db.models.deletion.CASCADE, to='identities.Identity')),
            ],
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.1 on 2016-02-04 14:27
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('identities', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='identity',
            name='communicate_through',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='identities_communicate_through', to='identities.Identity'),
        ),
        migrations.AddField(
            model_name='identity',
            name='operator',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='identities_created_by', to='identities.Identity'),
        ),
    ]



from django.conf.urls import url, include
from rest_framework import routers
from . import views

from seed_papertrail.decorators import papertrail

router = routers.DefaultRouter()
router.register(r'user', views.UserViewSet)
router.register(r'group', views.GroupViewSet)
router.register(r'identities', views.IdentityViewSet)
router.register(r'optout', views.OptOutViewSet)
router.register(r'optin', views.OptInViewSet)
router.register(r'webhook', views.HookViewSet)

# Wire up our API using automatic URL routing.
# Additionally, we include login URLs for the browseable API.
urlpatterns = [
    url(r'^api/v1/identities/search/$',
        papertrail.debug(sample=0.1)(views.IdentitySearchList.as_view())),
    url(r'^api/v1/identities/(?P<identity_id>.+)/addresses/(?P<address_type>.+)$',  # noqa
        papertrail.debug(sample=0.1)(views.IdentityAddresses.as_view())),
    url(r'^api/v1/user/token/$',
        papertrail.debug(sample=0.1)(views.UserView.as_view()),
        name='create-user-token'),
    url(r'^api/v1/detailkeys/',
        papertrail.debug(sample=0.1)(views.DetailKeyView.as_view())),
    # NOTE: Not sure how to wrap these router.urls in papertrail yet,
    #       seems like DRF makes this unnecessarily difficult.
    url(r'^api/v1/', include(router.urls)),
    url(r'^api/v1/optouts/search/$',
        papertrail.debug(sample=0.1)(views.OptOutSearchList.as_view())),
]



import json
import responses

try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse

from django.contrib.auth.models import User
from django.db.models.signals import post_save
from django.test import TestCase
from django.conf import settings
from rest_framework import status
from rest_framework.test import APIClient
from rest_framework.authtoken.models import Token
from rest_hooks.models import Hook
from requests_testadapter import TestAdapter, TestSession
from go_http.metrics import MetricsApiClient

from .models import (Identity, OptOut, OptIn, DetailKey, handle_optout,
                     handle_optin, fire_metrics_if_new, fire_optout_metric)
from .tasks import deliver_hook_wrapper, fire_metric, scheduled_metrics
from . import tasks


class RecordingAdapter(TestAdapter):

    """ Record the request that was handled by the adapter.
    """
    request = None

    def send(self, request, *args, **kw):
        self.request = request
        return super(RecordingAdapter, self).send(request, *args, **kw)


class APITestCase(TestCase):

    def setUp(self):
        self.client = APIClient()
        self.adminclient = APIClient()
        self.session = TestSession()


class AuthenticatedAPITestCase(APITestCase):

    def make_identity(self, id_data=None):
        if id_data is None:
            id_data = {
                "details": {
                    "name": "Test Name 1",
                    "default_addr_type": "msisdn",
                    "personnel_code": "12345",
                    "addresses": {
                        "msisdn": {
                            "+27123": {}
                        },
                        "email": {
                            "foo1@bar.com": {"default": True},
                            "foo2@bar.com": {}
                        }
                    }
                }
            }
        return Identity.objects.create(**id_data)

    def _replace_get_metric_client(self, session=None):
        return MetricsApiClient(
            auth_token=settings.METRICS_AUTH_TOKEN,
            api_url=settings.METRICS_URL,
            session=self.session)

    def _restore_get_metric_client(self, session=None):
        return MetricsApiClient(
            auth_token=settings.METRICS_AUTH_TOKEN,
            api_url=settings.METRICS_URL,
            session=session)

    def _replace_post_save_hooks(self):
        post_save.disconnect(handle_optout, sender=Identity)
        post_save.disconnect(handle_optin, sender=Identity)
        post_save.disconnect(fire_metrics_if_new, sender=Identity)
        post_save.disconnect(fire_optout_metric, sender=OptOut)

    def _restore_post_save_hooks(self):
        post_save.connect(handle_optout, sender=Identity)
        post_save.connect(handle_optin, sender=Identity)
        post_save.connect(fire_metrics_if_new, sender=Identity)
        post_save.connect(fire_optout_metric, sender=OptOut)

    def setUp(self):
        super(AuthenticatedAPITestCase, self).setUp()

        self._replace_post_save_hooks()
        tasks.get_metric_client = self._replace_get_metric_client

        self.username = 'testuser'
        self.password = 'testpass'
        self.user = User.objects.create_user(self.username,
                                             'testuser@example.com',
                                             self.password)
        token = Token.objects.create(user=self.user)
        self.token = token.key
        self.client.credentials(HTTP_AUTHORIZATION='Token ' + self.token)

        self.superuser = User.objects.create_superuser('testsu',
                                                       'su@example.com',
                                                       'dummypwd')
        sutoken = Token.objects.create(user=self.superuser)
        self.adminclient.credentials(
            HTTP_AUTHORIZATION='Token %s' % sutoken)

    def tearDown(self):
        self._restore_post_save_hooks()
        tasks.get_metric_client = self._restore_get_metric_client


class TestLogin(AuthenticatedAPITestCase):

    def test_login(self):
        # Setup
        post_auth = {"username": "testuser",
                     "password": "testpass"}
        # Execute
        request = self.client.post(
            '/api/token-auth/', post_auth)
        token = request.data.get('token', None)
        # Check
        self.assertIsNotNone(
            token, "Could not receive authentication token on login post.")
        self.assertEqual(
            request.status_code, 200,
            "Status code on /api/token-auth was %s (should be 200)."
            % request.status_code)


class TestUserCreation(AuthenticatedAPITestCase):

    def test_create_user_and_token(self):
        # Setup
        user_request = {"email": "test@example.org"}
        # Execute
        request = self.adminclient.post('/api/v1/user/token/', user_request)
        token = request.json().get('token', None)
        # Check
        self.assertIsNotNone(
            token, "Could not receive authentication token on post.")
        self.assertEqual(
            request.status_code, 201,
            "Status code on /api/v1/user/token/ was %s (should be 201)."
            % request.status_code)

    def test_create_user_and_token_fail_nonadmin(self):
        # Setup
        user_request = {"email": "test@example.org"}
        # Execute
        request = self.client.post('/api/v1/user/token/', user_request)
        error = request.json().get('detail', None)
        # Check
        self.assertIsNotNone(
            error, "Could not receive error on post.")
        self.assertEqual(
            error, "You do not have permission to perform this action.",
            "Error message was unexpected: %s."
            % error)

    def test_create_user_and_token_not_created(self):
        # Setup
        user_request = {"email": "test@example.org"}
        # Execute
        request = self.adminclient.post('/api/v1/user/token/', user_request)
        token = request.json().get('token', None)
        # And again, to get the same token
        request2 = self.adminclient.post('/api/v1/user/token/', user_request)
        token2 = request2.json().get('token', None)

        # Check
        self.assertEqual(
            token, token2,
            "Tokens are not equal, should be the same as not recreated.")

    def test_create_user_new_token_nonadmin(self):
        # Setup
        user_request = {"email": "test@example.org"}
        request = self.adminclient.post('/api/v1/user/token/', user_request)
        token = request.json().get('token', None)
        cleanclient = APIClient()
        cleanclient.credentials(HTTP_AUTHORIZATION='Token %s' % token)
        # Execute
        request = cleanclient.post('/api/v1/user/token/', user_request)
        error = request.json().get('detail', None)
        # Check
        # new user should not be admin
        self.assertIsNotNone(
            error, "Could not receive error on post.")
        self.assertEqual(
            error, "You do not have permission to perform this action.",
            "Error message was unexpected: %s."
            % error)


class TestIdentityAPI(AuthenticatedAPITestCase):

    def test_read_optouts(self):
        # Setup
        identity = self.make_identity()
        optout_data = {
            "request_source": "test_source",
            "requestor_source_id": "1",
            "address_type": "msisdn",
            "address": "+27123",
            "identity": identity,
            "reason": "not good messages",
        }
        OptOut.objects.create(**optout_data)
        # Execute
        response = self.client.get('/api/v1/optouts/search/',
                                   {"reason": "not good messages"},
                                   content_type='application/json')
        # Check
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data["results"]), 1)

    def test_read_optouts_invalid(self):
        # Setup
        identity = self.make_identity()
        optout_data = {
            "request_source": "test_source",
            "requestor_source_id": "1",
            "address_type": "msisdn",
            "address": "+27123",
            "identity": identity,
            "reason": "not good messages",
        }
        OptOut.objects.create(**optout_data)
        # Execute
        response = self.client.get('/api/v1/optouts/search/',
                                   {"foo": "bar"},
                                   content_type='application/json')
        # Check
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json()[0].replace("u'", "'"),
                         "Cannot resolve keyword 'foo' into field. Choices "
                         "are: address, address_type, created_at, created_by, "
                         "created_by_id, id, identity, identity_id, "
                         "optout_type, reason, request_source, "
                         "requestor_source_id")

    def test_read_optouts_filter(self):
        # Setup
        identity = self.make_identity()
        optout_data = {
            "request_source": "test_source",
            "requestor_source_id": "1",
            "address_type": "msisdn",
            "address": "+27123",
            "identity": identity,
            "reason": "not good messages",
        }
        OptOut.objects.create(**optout_data)
        # Execute
        response = self.client.get('/api/v1/optouts/search/',
                                   {"reason": "very good messages"},
                                   content_type='application/json')
        # Check
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data["results"]), 0)

    def test_read_identity(self):
        # Setup
        identity = self.make_identity()
        # Execute
        response = self.client.get('/api/v1/identities/%s/' % identity.id,
                                   content_type='application/json')
        # Check
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        d = Identity.objects.last()
        self.assertEqual(d.details["name"], "Test Name 1")
        self.assertEqual(d.version, 1)

    def test_read_identity_search_msisdn_single(self):
        # Setup
        self.make_identity()
        self.make_identity(id_data={
            "details": {
                "name": "Test Name 2",
                "default_addr_type": "msisdn",
                "personnel_code": "23456",
                "addresses": {
                    "msisdn": {
                        "+27123": {}
                    }
                }
            }
        })
        self.make_identity(id_data={
            "version": 2,
            "details": {
                "name": "Test Name 3",
                "addresses": {
                    "msisdn": {
                        "+27555": {}
                    }
                }
            }
        })
        # Execute
        response = self.client.get('/api/v1/identities/search/',
                                   {"details__addresses__msisdn": "+27555"},
                                   content_type='application/json')
        # Check
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data["results"]), 1)
        self.assertEqual(data["results"][0]["details"]["name"], "Test Name 3")

    def test_read_identity_search_msisdn_multiple(self):
        # Setup
        self.make_identity()
        self.make_identity(id_data={
            "details": {
                "name": "Test Name 2",
                "default_addr_type": "msisdn",
                "personnel_code": "23456",
                "addresses": {
                    "msisdn": {
                        "+27123": {}
                    }
                }
            }
        })
        self.make_identity(id_data={
            "version": 2,
            "details": {
                "name": "Test Name 3",
                "addresses": {
                    "msisdn": {
                        "+27555": {}
                    }
                }
            }
        })
        # Execute
        response = self.client.get('/api/v1/identities/search/',
                                   {"details__addresses__msisdn": "+27123"},
                                   content_type='application/json')
        # Check
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data["results"]), 2)

    def test_read_identity_search_msisdn_inactive_filter(self):
        # Setup
        self.make_identity(id_data={
            "details": {
                "name": "Test Name 2",
                "default_addr_type": "msisdn",
                "personnel_code": "23456",
                "addresses": {
                    "msisdn": {
                        "+27123": {"default": True, "inactive": True}
                    }
                }
            }
        })
        self.make_identity(id_data={
            "version": 2,
            "details": {
                "name": "Test Name 4",
                "addresses": {
                    "msisdn": {
                        "+27123": {}
                    }
                }
            }
        })
        # Execute
        response_default = self.client.get(
            '/api/v1/identities/search/',
            {
                "details__addresses__msisdn": "+27123"
            },
            content_type='application/json')
        response_include_inactive = self.client.get(
            '/api/v1/identities/search/',
            {
                "details__addresses__msisdn": "+27123",
                "include_inactive": True
            },
            content_type='application/json')
        response_exclude_inactive = self.client.get(
            '/api/v1/identities/search/',
            {
                "details__addresses__msisdn": "+27123",
                "include_inactive": False
            },
            content_type='application/json')
        # Check
        self.assertEqual(response_default.status_code,
                         status.HTTP_200_OK)
        data_default = response_default.json()
        self.assertEqual(len(data_default["results"]), 2)

        self.assertEqual(response_include_inactive.status_code,
                         status.HTTP_200_OK)
        data_include = response_include_inactive.json()
        self.assertEqual(len(data_include["results"]), 2)

        self.assertEqual(response_exclude_inactive.status_code,
                         status.HTTP_200_OK)
        data_exclude = response_exclude_inactive.json()
        self.assertEqual(len(data_exclude["results"]), 1)
        self.assertEqual(data_exclude["results"][0]["details"]["name"],
                         "Test Name 4")

    def test_read_identity_search_email(self):
        # Setup
        self.make_identity()
        self.make_identity(id_data={
            "details": {
                "name": "Test Name 2",
                "default_addr_type": "msisdn",
                "personnel_code": "23456",
                "addresses": {
                    "msisdn": {
                        "+27123": {}
                    }
                }
            }
        })
        self.make_identity(id_data={
            "version": 2,
            "details": {
                "name": "Test Name 3",
                "addresses": {
                    "msisdn": {
                        "+27555": {}
                    }
                }
            }
        })
        # Execute
        response = self.client.get(
            '/api/v1/identities/search/',
            {
                "details__addresses__email": "foo1@bar.com",
                "version": 1,
                "include_inactive": True
            },
            content_type='application/json')
        # Check
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data["results"]), 1)
        self.assertEqual(data["results"][0]["details"]["name"], "Test Name 1")

    def test_read_identity_search_msisdn_email_inactive_filter(self):
        # Setup
        self.make_identity(id_data={
            "details": {
                "name": "Test Name 2",
                "default_addr_type": "msisdn",
                "personnel_code": "23456",
                "addresses": {
                    "email": {
                        "foo@bar.com": {"default": True, "inactive": True}
                    }
                }
            }
        })
        self.make_identity(id_data={
            "version": 2,
            "details": {
                "name": "Test Name 4",
                "addresses": {
                    "email": {
                        "foo@bar.com": {}
                    }
                }
            }
        })
        # Execute
        response_include_inactive = self.client.get(
            '/api/v1/identities/search/',
            {
                "details__addresses__email": "foo@bar.com",
                "include_inactive": True
            },
            content_type='application/json')
        response_exclude_inactive = self.client.get(
            '/api/v1/identities/search/',
            {
                "details__addresses__email": "foo@bar.com",
                "include_inactive": False
            },
            content_type='application/json')
        # Check
        self.assertEqual(response_include_inactive.status_code,
                         status.HTTP_200_OK)
        data_include = response_include_inactive.json()
        self.assertEqual(len(data_include["results"]), 2)

        self.assertEqual(response_exclude_inactive.status_code,
                         status.HTTP_200_OK)
        data_exclude = response_exclude_inactive.json()
        self.assertEqual(len(data_exclude["results"]), 1)
        self.assertEqual(data_exclude["results"][0]["details"]["name"],
                         "Test Name 4")

    def test_read_identity_search_msisdn_email_and_msisdn(self):
        # Setup
        self.make_identity()
        self.make_identity(id_data={
            "details": {
                "name": "Test Name 2",
                "default_addr_type": "msisdn",
                "personnel_code": "23456",
                "addresses": {
                    "email": {
                        "foo@bar.com": {"default": True, "inactive": True}
                    },
                    "msisdn": {
                        "+27123": {"default": True}
                    }
                }
            }
        })
        self.make_identity(id_data={
            "version": 2,
            "details": {
                "name": "Test Name 4",
                "addresses": {
                    "email": {
                        "foo@bar.com": {}
                    },
                    "msisdn": {
                        "+27123": {}
                    }
                }
            }
        })
        self.make_identity(id_data={
            "version": 2,
            "details": {
                "name": "Test Name 5",
                "addresses": {
                    "email": {
                        "foo@bar.com": {}
                    },
                    "msisdn": {
                        "+27123": {"inactive": True}
                    }
                }
            }
        })
        # Execute
        response_include_inactive = self.client.get(
            '/api/v1/identities/search/',
            {
                "details__addresses__email": "foo@bar.com",
                "details__addresses__msisdn": "+27123",
                "include_inactive": True
            },
            content_type='application/json')
        response_exclude_inactive = self.client.get(
            '/api/v1/identities/search/',
            {
                "details__addresses__email": "foo@bar.com",
                "details__addresses__msisdn": "+27123",
                "include_inactive": False
            },
            content_type='application/json')
        # Check
        self.assertEqual(response_include_inactive.status_code,
                         status.HTTP_200_OK)
        data_include = response_include_inactive.json()
        self.assertEqual(len(data_include["results"]), 3)

        self.assertEqual(response_exclude_inactive.status_code,
                         status.HTTP_200_OK)
        data_exclude = response_exclude_inactive.json()
        self.assertEqual(len(data_exclude["results"]), 1)
        self.assertEqual(data_exclude["results"][0]["details"]["name"],
                         "Test Name 4")

    def test_read_identity_search_personnel_code(self):
        # Setup
        self.make_identity()
        self.make_identity({
            "details": {
                "name": "Test Name 2",
                "default_addr_type": "msisdn",
                "personnel_code": "23456",
                "addresses": {
                    "msisdn": {
                        "+27123": {}
                    }
                }
            }
        })
        self.make_identity(id_data={
            "version": 2,
            "details": {
                "name": "Test Name 3",
                "addresses": {
                    "msisdn": {
                        "+27555": {}
                    }
                }
            }
        })
        # Execute
        response = self.client.get('/api/v1/identities/search/',
                                   {"details__personnel_code": "23456"},
                                   content_type='application/json')
        # Check
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data["results"]), 1)
        self.assertEqual(data["results"][0]["details"]["name"], "Test Name 2")

    def test_read_identity_search_version(self):
        # Setup
        self.make_identity()
        self.make_identity(id_data={
            "details": {
                "name": "Test Name 2",
                "default_addr_type": "msisdn",
                "personnel_code": "23456",
                "addresses": {
                    "msisdn": {
                        "+27123": {}
                    }
                }
            }
        })
        self.make_identity(id_data={
            "version": 2,
            "details": {
                "name": "Test Name 3",
                "addresses": {
                    "msisdn": {
                        "+27555": {}
                    }
                }
            }
        })
        # Execute
        response = self.client.get('/api/v1/identities/search/',
                                   {"version": 2},
                                   content_type='application/json')
        # Check
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data["results"]), 1)
        self.assertEqual(data["results"][0]["details"]["name"], "Test Name 3")

    def test_read_identity_search_communicate_through(self):
        # Setup
        self.make_identity()
        test_id2 = self.make_identity(id_data={
            "details": {
                "name": "Test Name 2",
                "default_addr_type": "msisdn",
                "personnel_code": "23456",
                "addresses": {
                    "msisdn": {
                        "+27123": {}
                    }
                }
            }
        })
        test_id3 = {
            "version": 2,
            "details": {
                "name": "Test Name 3",
                "addresses": {
                    "msisdn": {
                        "+27555": {}
                    }
                }
            }
        }.copy()
        test_id3["communicate_through"] = test_id2
        self.make_identity(id_data=test_id3)
        # Execute
        response = self.client.get('/api/v1/identities/search/',
                                   {"communicate_through": str(test_id2.id)},
                                   content_type='application/json')
        # Check
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data["results"]), 1)
        self.assertEqual(data["results"][0]["details"]["name"], "Test Name 3")

    def test_read_identity_addresses_one_no_default(self):
        # Setup
        identity = self.make_identity(id_data={
            "details": {
                "name": "Test One No Default",
                "default_addr_type": "msisdn",
                "personnel_code": "23456",
                "addresses": {
                    "msisdn": {
                        "+27124": {}
                    }
                }
            }
        })
        # Execute
        response = self.client.get(
            '/api/v1/identities/%s/addresses/msisdn' % identity,
            {"default": "True"},
            content_type='application/json')
        # Check
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data["results"]), 1)
        self.assertEqual(data["results"][0]["address"], "+27124")

    def test_read_identity_addresses_two_with_default(self):
        # Setup
        identity = self.make_identity(id_data={
            "details": {
                "name": "Test One No Default",
                "default_addr_type": "msisdn",
                "personnel_code": "23456",
                "addresses": {
                    "msisdn": {
                        "+27124": {},
                        "+27125": {"default": True}
                    }
                }
            }
        })
        # Execute
        response = self.client.get(
            '/api/v1/identities/%s/addresses/msisdn' % identity,
            {"default": "True"},
            content_type='application/json')
        # Check
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data["results"]), 1)
        self.assertEqual(data["results"][0]["address"], "+27125")

    def test_read_identity_addresses_two_with_optout(self):
        # Setup
        identity = self.make_identity(id_data={
            "details": {
                "name": "Test One No Default",
                "default_addr_type": "msisdn",
                "personnel_code": "23456",
                "addresses": {
                    "msisdn": {
                        "+27124": {},
                        "+27125": {"default": True, "optedout": True}
                    }
                }
            }
        })
        # Execute
        response = self.client.get(
            '/api/v1/identities/%s/addresses/msisdn' % identity,
            {"default": "True"},
            content_type='application/json')
        # Check
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        # default address is marked as optedout
        self.assertEqual(len(data["results"]), 0)

    def test_read_identity_addresses_with_communicate_through(self):
        # Setup
        identity = self.make_identity(id_data={
            "details": {
                "name": "Test One No Default",
                "default_addr_type": "msisdn",
                "personnel_code": "23456",
                "addresses": {
                    "msisdn": {
                        "+27124": {},
                        "+27125": {"default": True}
                    }
                }
            }
        })
        sub_identity = self.make_identity(id_data={
            "communicate_through": identity,
            "details": {
                "name": "Test One CT",
                "default_addr_type": "msisdn",
                "personnel_code": "23457",
                "addresses": {
                    "msisdn": {
                        "+27321": {},
                        "+27543": {"default": True}
                    }
                }
            }
        })
        # Execute
        response = self.client.get(
            '/api/v1/identities/%s/addresses/msisdn' % sub_identity,
            {"default": "True", "use_communicate_through": "True"},
            content_type='application/json')
        # Check
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data["results"]), 1)
        self.assertEqual(data["results"][0]["address"], "+27125")

    def test_read_identity_addresses_without_communicate_through(self):
        # Setup
        identity = self.make_identity(id_data={
            "details": {
                "name": "Test One No Default",
                "default_addr_type": "msisdn",
                "personnel_code": "23456",
                "addresses": {
                    "msisdn": {
                        "+27124": {},
                        "+27125": {"default": True}
                    }
                }
            }
        })
        sub_identity = self.make_identity(id_data={
            "communicate_through": identity,
            "details": {
                "name": "Test One CT",
                "default_addr_type": "msisdn",
                "personnel_code": "23457",
                "addresses": {
                    "msisdn": {
                        "+27321": {},
                        "+27543": {"default": True}
                    }
                }
            }
        })
        # Execute
        response = self.client.get(
            '/api/v1/identities/%s/addresses/msisdn' % sub_identity,
            {"default": "True"},
            content_type='application/json')
        # Check
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data["results"]), 1)
        self.assertEqual(data["results"][0]["address"], "+27543")

    def test_update_identity(self):
        # Setup
        identity = self.make_identity()
        new_details = {
            "details": {
                "name": "Changed Name",
                "default_addr_type": "email",
                "addresses": {
                    "msisdn": {"+27123": {}},
                    "email": {"foo1@bar.com": {}, "foo2@bar.com": {}}
                }
            }
        }
        # Execute
        response = self.client.patch('/api/v1/identities/%s/' % identity.id,
                                     json.dumps(new_details),
                                     content_type='application/json')
        # Check
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        d = Identity.objects.last()
        self.assertEqual(d.details["name"], "Changed Name")
        self.assertEqual(d.version, 1)

    def test_delete_identity(self):
        # Setup
        identity = self.make_identity()
        # Execute
        response = self.client.delete('/api/v1/identities/%s/' % identity.id,
                                      content_type='application/json')
        # Check
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        d = Identity.objects.filter().count()
        self.assertEqual(d, 0)

    def test_create_identity(self):
        # Setup
        identity1 = self.make_identity()
        identity2 = self.make_identity(id_data={
            "details": {
                "name": "Test Name 2",
                "default_addr_type": "msisdn",
                "personnel_code": "23456",
                "addresses": {
                    "msisdn": {
                        "+27123": {}
                    }
                }
            }
        })
        post_identity = {
            "communicate_through": str(identity1.id),
            "operator": str(identity2.id),
            "details": {
                "name": "Test Name",
                "default_addr_type": "msisdn",
                "addresses": "msisdn:+27123 email:foo@bar.com"
            }
        }
        # Execute
        response = self.client.post('/api/v1/identities/',
                                    json.dumps(post_identity),
                                    content_type='application/json')
        # Check
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        d = Identity.objects.get(id=response.data["id"])
        self.assertEqual(d.details["name"], "Test Name")
        self.assertEqual(d.version, 1)

    def test_create_identity_no_details(self):
        # Setup
        post_identity = {
            "details": {}
        }
        # Execute
        response = self.client.post('/api/v1/identities/',
                                    json.dumps(post_identity),
                                    content_type='application/json')
        # Check
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        d = Identity.objects.last()
        self.assertEqual(d.version, 1)

    def test_create_identity_detailkeys(self):
        # Setup
        self.make_identity()

        # Check
        c = DetailKey.objects.all().count()
        self.assertEqual(c, 4)

    def test_create_identity_detailkeys_two_new(self):
        # Setup
        self.make_identity()
        self.make_identity(id_data={
            "details": {
                "fresh": "as",
                "default_addr_type": "msisdn",
                "a": "daisy"
            }
        })

        # Check
        c = DetailKey.objects.all().count()
        self.assertEqual(c, 6)

    def test_identity_detailkeys_view(self):
        # Setup
        self.make_identity()

        # Execute
        response = self.client.get('/api/v1/detailkeys/',
                                   content_type='application/json')
        # Check
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data["key_names"]), 4)
        self.assertEqual("default_addr_type" in data["key_names"], True)


class TestOptInAPI(AuthenticatedAPITestCase):
    def test_create_optin_with_identity(self):
        # Setup
        identity = self.make_identity()
        optin_data = {
            "request_source": "test_source",
            "requestor_source_id": "1",
            "address_type": "msisdn",
            "address": "+27123",
            "identity": str(identity.id)
        }
        # Execute
        response = self.client.post('/api/v1/optin/',
                                    json.dumps(optin_data),
                                    content_type='application/json')
        # Check
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        d = OptIn.objects.get(id=response.data["id"])
        self.assertEqual(d.identity, identity)
        self.assertEqual(d.request_source, "test_source")
        self.assertEqual(d.requestor_source_id, '1')

    def test_create_optin_with_address(self):
        # Setup
        identity = self.make_identity()
        optin_data = {
            "request_source": "test_source",
            "requestor_source_id": "1",
            "address_type": "msisdn",
            "address": "+27123",
        }
        # Execute
        response = self.client.post('/api/v1/optin/',
                                    json.dumps(optin_data),
                                    content_type='application/json')
        # Check
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        d = OptIn.objects.get(id=response.data["id"])
        self.assertEqual(d.identity, identity)
        self.assertEqual(d.request_source, "test_source")
        self.assertEqual(d.requestor_source_id, '1')

    def test_create_optin_no_matching_address(self):
        # Setup
        optin_data = {
            "request_source": "test_source",
            "requestor_source_id": "1",
            "address_type": "msisdn",
            "address": "+27123"
        }
        # Execute
        response = self.client.post('/api/v1/optin/',
                                    json.dumps(optin_data),
                                    content_type='application/json')
        # Check
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json()[0],
                         "There is no identity with this address.")

    def test_create_optin_multiple_matching_addresses(self):
        # Setup
        self.make_identity()
        self.make_identity()
        optin_data = {
            "request_source": "test_source",
            "requestor_source_id": "1",
            "address_type": "msisdn",
            "address": "+27123",
        }
        # Execute
        response = self.client.post('/api/v1/optin/',
                                    json.dumps(optin_data),
                                    content_type='application/json')
        # Check
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(
            response.json()[0],
            "There are multiple identities with this address.")

    def test_create_webhook(self):
        # Setup
        user = User.objects.get(username='testuser')
        post_data = {
            "target": "http://example.com/test_source/",
            "event": "optin.requested"
        }
        # Execute
        response = self.client.post('/api/v1/webhook/',
                                    json.dumps(post_data),
                                    content_type='application/json')
        # Check
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        d = Hook.objects.last()
        self.assertEqual(d.target, 'http://example.com/test_source/')
        self.assertEqual(d.user, user)

    @responses.activate
    def test_deliver_hook_task(self):
        # Setup
        user = User.objects.get(username='testuser')
        hook = Hook.objects.create(
            user=user,
            event='optin.requested',
            target='http://example.com/api/v1/')
        payload = {
            "identity": "test-219f0f88-7d2b-414d-933c-1f8e652869c4",
            "identity_details": {
                "addresses": {
                    "msisdn": {
                        "+27123": {"optedout": True}
                    }
                }
            },
            "optin_address_type": "msisdn",
            "optin_address": "+27123"
        }
        responses.add(
            responses.POST,
            'http://example.com/api/v1/',
            json.dumps(payload),
            status=200, content_type='application/json')

        deliver_hook_wrapper('http://example.com/api/v1/', payload, None, hook)

        # Execute
        self.assertEqual(responses.calls[0].request.url,
                         "http://example.com/api/v1/")

    @responses.activate
    def test_optin(self):
        # Setup
        post_save.connect(receiver=handle_optin, sender=OptIn)
        user = User.objects.get(username='testuser')
        Hook.objects.create(user=user,
                            event='optin.requested',
                            target='http://example.com/api/v1/')
        identity = self.make_identity()
        identity.details["addresses"]["msisdn"]["+27123"]["optedout"] = True
        identity.save()

        payload = {
            "identity": str(identity.id),
            "identity_details": identity.details,
            "optin_address_type": "msisdn",
            "optin_address": "+27123"
        }
        responses.add(
            responses.POST,
            'http://example.com/api/v1/',
            json.dumps(payload),
            status=200, content_type='application/json')

        OptIn.objects.create(
            identity=identity, created_by=user, request_source="test_source",
            requestor_source_id=1, address_type="msisdn", address="+27123")

        self.assertEqual(responses.calls[0].request.url,
                         'http://example.com/api/v1/')
        identity = Identity.objects.get(pk=identity.pk)
        self.assertEqual(identity.details, {
            "name": "Test Name 1",
            "default_addr_type": "msisdn",
            "personnel_code": "12345",
            "addresses": {
                "msisdn": {
                    "+27123": {"optedout": False}
                },
                "email": {
                    "foo1@bar.com": {"default": True},
                    "foo2@bar.com": {}
                }
            }
        })


class TestOptOutAPI(AuthenticatedAPITestCase):
    def test_create_optout_with_identity(self):
        # Setup
        identity = self.make_identity()
        optout_data = {
            "request_source": "test_source",
            "requestor_source_id": "1",
            "address_type": "msisdn",
            "address": "+27123",
            "identity": str(identity.id)
        }
        # Execute
        response = self.client.post('/api/v1/optout/',
                                    json.dumps(optout_data),
                                    content_type='application/json')
        # Check
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        d = OptOut.objects.get(id=response.data["id"])
        self.assertEqual(d.identity, identity)
        self.assertEqual(d.request_source, "test_source")
        self.assertEqual(d.requestor_source_id, '1')

    def test_create_optout_with_address(self):
        # Setup
        identity = self.make_identity()
        optout_data = {
            "request_source": "test_source",
            "requestor_source_id": "1",
            "address_type": "msisdn",
            "address": "+27123",
            "reason": "not good messages"
        }
        # Execute
        response = self.client.post('/api/v1/optout/',
                                    json.dumps(optout_data),
                                    content_type='application/json')
        # Check
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        d = OptOut.objects.get(id=response.data["id"])
        self.assertEqual(d.identity, identity)
        self.assertEqual(d.request_source, "test_source")
        self.assertEqual(d.requestor_source_id, '1')
        self.assertEqual(d.reason, 'not good messages')

    def test_create_optout_no_matching_address(self):
        # Setup
        optout_data = {
            "request_source": "test_source",
            "requestor_source_id": "1",
            "address_type": "msisdn",
            "address": "+27123"
        }
        # Execute
        response = self.client.post('/api/v1/optout/',
                                    json.dumps(optout_data),
                                    content_type='application/json')
        # Check
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json()[0],
                         "There is no identity with this address.")

    def test_create_opt_out_multiple_matching_addresses(self):
        # Setup
        self.make_identity()
        self.make_identity()
        optout_data = {
            "request_source": "test_source",
            "requestor_source_id": "1",
            "address_type": "msisdn",
            "address": "+27123",
            "optout_type": "forget"
        }
        # Execute
        response = self.client.post('/api/v1/optout/',
                                    json.dumps(optout_data),
                                    content_type='application/json')
        # Check
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(
            response.json()[0],
            "There are multiple identities with this address.")

    def test_create_webhook(self):
        # Setup
        user = User.objects.get(username='testuser')
        post_data = {
            "target": "http://example.com/test_source/",
            "event": "optout.requested"
        }
        # Execute
        response = self.client.post('/api/v1/webhook/',
                                    json.dumps(post_data),
                                    content_type='application/json')
        # Check
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        d = Hook.objects.last()
        self.assertEqual(d.target, 'http://example.com/test_source/')
        self.assertEqual(d.user, user)

    @responses.activate
    def test_deliver_hook_task(self):
        # Setup
        user = User.objects.get(username='testuser')
        hook = Hook.objects.create(
            user=user,
            event='optout.requested',
            target='http://example.com/api/v1/')
        payload = {
            "identity": "test-219f0f88-7d2b-414d-933c-1f8e652869c4",
            "identity_details": {
                "addresses": {
                    "msisdn": {
                        "+27123": {}
                    }
                }
            },
            "optout_type": "forget"
        }
        responses.add(
            responses.POST,
            'http://example.com/api/v1/',
            json.dumps(payload),
            status=200, content_type='application/json')

        deliver_hook_wrapper('http://example.com/api/v1/', payload, None, hook)

        # Execute
        self.assertEqual(responses.calls[0].request.url,
                         "http://example.com/api/v1/")

    @responses.activate
    def test_optout_webhook_combination(self):
        # Setup
        post_save.connect(receiver=handle_optout, sender=OptOut)
        user = User.objects.get(username='testuser')
        Hook.objects.create(user=user,
                            event='optout.requested',
                            target='http://example.com/api/v1/')
        identity = self.make_identity()
        payload = {
            "identity": str(identity.id),
            "identity_details": identity.details,
            'optout_type': "forget",
        }
        responses.add(
            responses.POST,
            'http://example.com/api/v1/',
            json.dumps(payload),
            status=200, content_type='application/json')

        OptOut.objects.create(
            identity=identity, created_by=user, request_source="test_source",
            requestor_source_id=1, address_type="msisdn", address="+27123",
            optout_type="forget")

        self.assertEqual(responses.calls[0].request.url,
                         'http://example.com/api/v1/')
        identity = Identity.objects.get(pk=identity.pk)
        self.assertEqual(identity.details, {
            "name": "redacted",
            "default_addr_type": "redacted",
            "personnel_code": "redacted",
            "addresses": {}
        })

    @responses.activate
    def test_optout_webhook_stop(self):
        # Setup
        post_save.connect(receiver=handle_optout, sender=OptOut)
        user = User.objects.get(username='testuser')
        Hook.objects.create(user=user,
                            event='optout.requested',
                            target='http://example.com/api/v1/')
        identity = self.make_identity()
        payload = {
            "identity": str(identity.id),
            "identity_details": identity.details,
            'optout_type': "stop",
        }
        responses.add(
            responses.POST,
            'http://example.com/api/v1/',
            json.dumps(payload),
            status=200, content_type='application/json')

        OptOut.objects.create(
            identity=identity, created_by=user, request_source="test_source",
            requestor_source_id=1, address_type="msisdn", address="+27123",
            optout_type="stop")

        self.assertEqual(responses.calls[0].request.url,
                         'http://example.com/api/v1/')
        identity = Identity.objects.get(pk=identity.pk)
        self.assertEqual(identity.details, {
            "name": "Test Name 1",
            "default_addr_type": "msisdn",
            "personnel_code": "12345",
            "addresses": {
                "msisdn": {
                    "+27123": {"optedout": True}
                },
                "email": {
                    "foo1@bar.com": {"default": True},
                    "foo2@bar.com": {}
                }
            }
        })

    @responses.activate
    def test_optout_webhook_stop_all(self):
        # Setup
        post_save.connect(receiver=handle_optout, sender=OptOut)
        user = User.objects.get(username='testuser')
        Hook.objects.create(user=user,
                            event='optout.requested',
                            target='http://example.com/api/v1/')
        identity = self.make_identity()
        payload = {
            "identity": str(identity.id),
            "identity_details": identity.details,
            'optout_type': "stopall",
        }
        responses.add(
            responses.POST,
            'http://example.com/api/v1/',
            json.dumps(payload),
            status=200, content_type='application/json')

        OptOut.objects.create(
            identity=identity, created_by=user, request_source="test_source",
            requestor_source_id=1, address_type="msisdn", address="+27123",
            optout_type="stopall")

        self.assertEqual(responses.calls[0].request.url,
                         'http://example.com/api/v1/')
        identity = Identity.objects.get(pk=identity.pk)
        self.assertEqual(identity.details, {
            "name": "Test Name 1",
            "default_addr_type": "msisdn",
            "personnel_code": "12345",
            "addresses": {
                "msisdn": {
                    "+27123": {"optedout": True}
                },
                "email": {
                    "foo1@bar.com": {"default": True, "optedout": True},
                    "foo2@bar.com": {"optedout": True}
                }
            }
        })


class TestHealthcheckAPI(AuthenticatedAPITestCase):

    def test_healthcheck_read(self):
        # Setup
        # Execute
        response = self.client.get('/api/health/',
                                   content_type='application/json')
        # Check
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data["up"], True)
        self.assertEqual(response.data["result"]["database"], "Accessible")


class TestMetricsAPI(AuthenticatedAPITestCase):

    def test_metrics_read(self):
        # Setup
        # Execute
        response = self.client.get('/api/metrics/',
                                   content_type='application/json')
        # Check
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            sorted(response.data["metrics_available"]), sorted([
                'identities.created.sum',
                'identities.created.last',
                'identities.change.msisdn.sum',
                'identities.change.email.sum',
                'optout.sum',
                'optout.total.last',
            ])
        )

    @responses.activate
    def test_post_metrics(self):
        # Setup
        # deactivate Testsession for this test
        self.session = None
        responses.add(responses.POST,
                      "http://metrics-url/metrics/",
                      json={"foo": "bar"},
                      status=200, content_type='application/json')
        # Execute
        response = self.client.post('/api/metrics/',
                                    content_type='application/json')
        # Check
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response.data["scheduled_metrics_initiated"], True)


class TestMetrics(AuthenticatedAPITestCase):

    def check_request(
            self, request, method, params=None, data=None, headers=None):
        self.assertEqual(request.method, method)
        if params is not None:
            url = urlparse.urlparse(request.url)
            qs = urlparse.parse_qsl(url.query)
            self.assertEqual(dict(qs), params)
        if headers is not None:
            for key, value in headers.items():
                self.assertEqual(request.headers[key], value)
        if data is None:
            self.assertEqual(request.body, None)
        else:
            self.assertEqual(json.loads(request.body), data)

    def _mount_session(self):
        response = [{
            'name': 'foo',
            'value': 9000,
            'aggregator': 'bar',
        }]
        adapter = RecordingAdapter(json.dumps(response).encode('utf-8'))
        self.session.mount(
            "http://metrics-url/metrics/", adapter)
        return adapter

    def test_direct_fire(self):
        # Setup
        adapter = self._mount_session()
        # Execute
        result = fire_metric.apply_async(kwargs={
            "metric_name": 'foo.last',
            "metric_value": 1,
            "session": self.session
        })
        # Check
        self.check_request(
            adapter.request, 'POST',
            data={"foo.last": 1.0}
        )
        self.assertEqual(result.get(),
                         "Fired metric <foo.last> with value <1.0>")

    def test_created_metrics(self):
        # Setup
        adapter = self._mount_session()
        # reconnect metric post_save hook
        post_save.connect(fire_metrics_if_new, sender=Identity)

        # Execute
        self.make_identity()

        # Check
        self.check_request(
            adapter.request, 'POST',
            data={"identities.created.sum": 1.0}
        )
        # remove post_save hooks to prevent teardown errors
        post_save.disconnect(fire_metrics_if_new, sender=Identity)

    @responses.activate
    def test_optout_created_metrics(self):
        # Setup
        identity = self.make_identity()
        optout_data = {
            "request_source": "test_source",
            "requestor_source_id": "1",
            "address_type": "msisdn",
            "address": "+27123",
            "identity": identity,
            "reason": "not good messages",
        }
        OptOut.objects.create(**optout_data)

        self.session = None
        # reconnect metric post_save hook
        post_save.connect(fire_optout_metric, sender=OptOut)

        responses.add(responses.POST,
                      "http://metrics-url/metrics/",
                      json={"foo": "bar"},
                      status=200, content_type='application/json')

        # Execute
        OptOut.objects.create(**optout_data)

        # Check
        self.assertEqual(len(responses.calls), 2)
        [call1, call2] = responses.calls
        self.assertEqual(json.loads(call1.request.body), {"optout.sum": 1.0})
        self.assertEqual(json.loads(call2.request.body),
                         {"optout.total.last": 2.0})
        # remove post_save hooks to prevent teardown errors
        post_save.disconnect(fire_optout_metric, sender=OptOut)

    @responses.activate
    def test_multiple_created_metrics(self):
        # Setup
        # deactivate Testsession for this test
        self.session = None
        # reconnect metric post_save hook
        post_save.connect(fire_metrics_if_new, sender=Identity)
        # add metric post response
        responses.add(responses.POST,
                      "http://metrics-url/metrics/",
                      json={"foo": "bar"},
                      status=200, content_type='application/json')

        # Execute
        self.make_identity()
        self.make_identity()

        # Check
        self.assertEqual(len(responses.calls), 2)
        # remove post_save hooks to prevent teardown errors
        post_save.disconnect(fire_metrics_if_new, sender=Identity)

    @responses.activate
    def test_scheduled_metrics(self):
        # Setup
        # deactivate Testsession for this test
        self.session = None
        # add metric post response
        responses.add(responses.POST,
                      "http://metrics-url/metrics/",
                      json={"foo": "bar"},
                      status=200, content_type='application/json')

        # Execute
        result = scheduled_metrics.apply_async()
        # Check
        self.assertEqual(result.get(), "1 Scheduled metrics launched")
        # fire_messagesets_tasks fires two metrics, therefore extra call
        self.assertEqual(len(responses.calls), 1)

    def test_fire_created_last(self):
        # Setup
        adapter = self._mount_session()
        # make two identities
        self.make_identity()
        self.make_identity()

        # Execute
        result = tasks.fire_created_last.apply_async()

        # Check
        self.assertEqual(
            result.get().get(),
            "Fired metric <identities.created.last> with value <2.0>"
        )
        self.check_request(
            adapter.request, 'POST',
            data={"identities.created.last": 2.0}
        )

    def test_update_msisdn_metric(self):
        """
        When a MSISDN is changed on a identity, a sum metric should be fired
        """
        # Setup
        adapter = self._mount_session()

        identity = self.make_identity()
        new_details = {
            "addresses": {
                "msisdn": {"+27123999": {}},
                "email": {"foo1@bar.com": {}, "foo2@bar.com": {}}
            }
        }

        # Execute
        identity.details = new_details
        identity.save()

        # Check
        self.check_request(
            adapter.request, 'POST',
            data={"identities.change.msisdn.sum": 1.0}
        )

    def test_update_msisdn_metric_negative(self):
        """
        When the details of the identity is updated and the MSISDN remains the
        same, no metric request should be made
        """
        # Setup
        adapter = self._mount_session()

        identity = self.make_identity()
        new_details = {
            "addresses": {
                "msisdn": {"+27123": {}},
                "email": {"foo1@bar.com": {}, "foo2@bar.com": {}}
            }
        }

        # Execute
        identity.details = new_details
        identity.save()

        # Check
        self.assertEqual(adapter.request, None)

    def test_update_email_metric(self):
        """
        When a email is changed on a identity, a sum metric should be fired
        """
        # Setup
        adapter = self._mount_session()

        identity = self.make_identity()
        new_details = {
            "addresses": {
                "msisdn": {"+27123": {}},
                "email": {"foo1_new@bar.com": {}, "foo2@bar.com": {}}
            }
        }

        # Execute
        identity.details = new_details
        identity.save()

        # Check
        self.check_request(
            adapter.request, 'POST',
            data={"identities.change.email.sum": 1.0}
        )



"""
Django settings for seed_identity_store project.

For more information on this file, see
https://docs.djangoproject.com/en/1.9/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/1.9/ref/settings/
"""

from kombu import Exchange, Queue

import os
import djcelery
import dj_database_url

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.dirname(__file__))

# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/1.9/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = os.environ.get('SECRET_KEY', 'REPLACEME')

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = os.environ.get('DEBUG', 'false').lower() == 'true'

ALLOWED_HOSTS = ['*']


# Application definition

INSTALLED_APPS = (
    # admin
    'django.contrib.admin',
    # core
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    # documentation
    'rest_framework_docs',
    # 3rd party
    'raven.contrib.django.raven_compat',
    'rest_framework',
    'rest_framework.authtoken',
    'django_filters',
    'rest_hooks',
    'djcelery',
    # us
    'identities',

)

MIDDLEWARE_CLASSES = (
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.auth.middleware.SessionAuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
)

ROOT_URLCONF = 'seed_identity_store.urls'

WSGI_APPLICATION = 'seed_identity_store.wsgi.application'


# Database
# https://docs.djangoproject.com/en/1.9/ref/settings/#databases

DATABASES = {
    'default': dj_database_url.config(
        default=os.environ.get(
            'IDENTITIES_DATABASE',
            'postgres://postgres:@localhost/seed_identity_store')),
}


# Internationalization
# https://docs.djangoproject.com/en/1.9/topics/i18n/

LANGUAGE_CODE = 'en-gb'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/1.9/howto/static-files/

STATICFILES_FINDERS = (
    'django.contrib.staticfiles.finders.AppDirectoriesFinder',
    'django.contrib.staticfiles.finders.FileSystemFinder',
)

STATIC_ROOT = 'staticfiles'
STATIC_URL = '/static/'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [
            os.path.join(BASE_DIR, 'templates')
        ],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

# Sentry configuration
RAVEN_CONFIG = {
    # DevOps will supply you with this.
    'dsn': os.environ.get('IDENTITIES_SENTRY_DSN', None),
}

# REST Framework conf defaults
REST_FRAMEWORK = {
    'PAGE_SIZE': 1000,
    'DEFAULT_PAGINATION_CLASS':
        'rest_framework.pagination.LimitOffsetPagination',
    'DEFAULT_AUTHENTICATION_CLASSES': (
        'rest_framework.authentication.BasicAuthentication',
        'rest_framework.authentication.TokenAuthentication',
    ),
    'DEFAULT_PERMISSION_CLASSES': (
        'rest_framework.permissions.IsAuthenticated',
    ),
    'DEFAULT_FILTER_BACKENDS': ('rest_framework.filters.DjangoFilterBackend',)
}

# Webhook event definition
HOOK_EVENTS = {
    # 'any.event.name': 'App.Model.Action' (created/updated/deleted)
    'optout.requested': None,
    'optin.requested': None,
    'identity.created': 'identities.Identity.created+'
}

HOOK_DELIVERER = 'identities.tasks.deliver_hook_wrapper'

HOOK_AUTH_TOKEN = os.environ.get('HOOK_AUTH_TOKEN', 'REPLACEME')

# Celery configuration options
CELERY_RESULT_BACKEND = 'djcelery.backends.database:DatabaseBackend'
CELERYBEAT_SCHEDULER = 'djcelery.schedulers.DatabaseScheduler'

BROKER_URL = os.environ.get('BROKER_URL', 'redis://localhost:6379/0')

CELERY_DEFAULT_QUEUE = 'seed_identity_store'
CELERY_QUEUES = (
    Queue('seed_identity_store',
          Exchange('seed_identity_store'),
          routing_key='seed_identity_store'),
)

CELERY_ALWAYS_EAGER = False

# Tell Celery where to find the tasks
CELERY_IMPORTS = (
    'identities.tasks',
)

CELERY_CREATE_MISSING_QUEUES = True
CELERY_ROUTES = {
    'celery.backend_cleanup': {
        'queue': 'mediumpriority',
    },
    'identities.tasks.DeliverHook': {
        'queue': 'priority',
    },
    'identities.tasks.fire_metric': {
        'queue': 'metrics',
    },
    'identities.tasks.scheduled_metrics': {
        'queue': 'metrics',
    },
    'identities.tasks.fire_active_last': {
        'queue': 'metrics',
    },
    'identities.tasks.populate_detail_key': {
        'queue': 'priority'
    }
}

ADDRESS_TYPES = ['msisdn', 'email']

METRICS_REALTIME = [
    'identities.created.sum',
    'optout.sum',
    'optout.total.last',
]

METRICS_REALTIME.extend(
    ['identities.change.%s.sum' % at for at in ADDRESS_TYPES])

METRICS_SCHEDULED = [
    'identities.created.last'
]
METRICS_SCHEDULED_TASKS = [
    'fire_created_last'
]

CELERY_TASK_SERIALIZER = 'json'
CELERY_RESULT_SERIALIZER = 'json'
CELERY_ACCEPT_CONTENT = ['json']
CELERY_IGNORE_RESULT = True
CELERYD_MAX_TASKS_PER_CHILD = 50

djcelery.setup_loader()

METRICS_URL = os.environ.get("METRICS_URL", None)
METRICS_AUTH_TOKEN = os.environ.get("METRICS_AUTH_TOKEN", "REPLACEME")


PAPERTRAIL = os.environ.get('PAPERTRAIL')
if PAPERTRAIL:
    import seed_papertrail  # noqa
    PAPERTRAIL_HOST, _, PAPERTRAIL_PORT = PAPERTRAIL.partition(':')
    LOGGING = seed_papertrail.auto_configure(
        host=PAPERTRAIL_HOST,
        port=int(PAPERTRAIL_PORT),
        system=os.environ.get('MARATHON_APP_DOCKER_IMAGE', 'seed'),
        program=os.environ.get('MESOS_TASK_ID', 'identity_store'))



# -*- coding: utf-8 -*-
# Generated by Django 1.9.1 on 2016-08-21 13:01
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('identities', '0003_optout'),
    ]

    operations = [
        migrations.CreateModel(
            name='DetailKey',
            fields=[
                ('key_name', models.CharField(max_length=200, primary_key=True, serialize=False)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
            ],
        ),
    ]



from __future__ import absolute_import

import os

import celery
import raven
from raven.contrib.celery import register_signal, register_logger_signal

from django.conf import settings

# set the default Django settings module for the 'celery' program.
os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'seed_identity_store.settings')


class Celery(celery.Celery):

    def on_configure(self):
        client = raven.Client(settings.RAVEN_CONFIG['dsn'])
        # register a custom filter to filter out duplicate logs
        register_logger_signal(client)
        # hook into the Celery error handler
        register_signal(client)


app = Celery('seed_identity_store')

# Using a string here means the worker will not have to
# pickle the object when using Windows.
app.config_from_object('django.conf:settings')
app.autodiscover_tasks(lambda: settings.INSTALLED_APPS)


@app.task(bind=True)
def debug_task(self):
    print('Request: {0!r}'.format(self.request))



from seed_identity_store.settings import *  # flake8: noqa

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = 'TESTSEKRET'

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

TEMPLATE_DEBUG = True

CELERY_EAGER_PROPAGATES_EXCEPTIONS = True
CELERY_ALWAYS_EAGER = True
BROKER_BACKEND = 'memory'
CELERY_RESULT_BACKEND = 'djcelery.backends.database:DatabaseBackend'

METRICS_URL = "http://metrics-url"
METRICS_AUTH_TOKEN = "REPLACEME"



# -*- coding: utf-8 -*-
# Generated by Django 1.9.1 on 2016-01-27 10:10
from __future__ import unicode_literals

from django.conf import settings
import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Identity',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('version', models.IntegerField(default=1)),
                ('details', django.contrib.postgres.fields.jsonb.JSONField()),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('created_by', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='identities_created', to=settings.AUTH_USER_MODEL)),
                ('updated_by', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='identities_updated', to=settings.AUTH_USER_MODEL)),
            ],
        ),
    ]



import uuid

from django.contrib.postgres.fields import JSONField
from django.contrib.auth.models import User
from django.db import models
from django.db.models.signals import post_save, pre_save
from django.dispatch import receiver
from django.utils.encoding import python_2_unicode_compatible
from django.core.exceptions import ValidationError
from django.core.cache import cache
from rest_hooks.signals import raw_hook_event

from seed_identity_store.settings import ADDRESS_TYPES


class IdentityManager(models.Manager):

    def filter_by_addr(self, address_type, address):
        filter_string = "details__addresses__%s__has_key" % (address_type, )
        return self.filter(**{filter_string: address})


@python_2_unicode_compatible
class Identity(models.Model):

    """
    version: 1
    details should contain at minimum:
    addresses: structured like -
        "addresses": {
            "msisdn": {
                "+27123": {}
            },
            "email": {
                "foo1@bar.com": {"default": True},
                "foo2@bar.com": {"optedout": True}
            }
        }
    default_addr_type: which addr_type in addresses to default to if non-given
    """
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    version = models.IntegerField(default=1)
    details = JSONField()
    communicate_through = models.ForeignKey(
        'self', related_name='identities_communicate_through',
        null=True, blank=True)
    operator = models.ForeignKey(
        'self', related_name='identities_created_by',
        null=True, blank=True)
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)
    created_by = models.ForeignKey(User, related_name='identities_created',
                                   null=True)
    updated_by = models.ForeignKey(User, related_name='identities_updated',
                                   null=True)
    user = property(lambda self: self.created_by)

    objects = IdentityManager()

    def serialize_hook(self, hook):
        return {
            'hook': hook.dict(),
            'data': {
                'id': str(self.id),
                'version': self.version,
                'details': self.details,
                'communicate_through': str(self.communicate_through),
                'operator': str(self.operator),
                'created_at': self.created_at.isoformat(),
                'created_by': self.created_by.username,
                'updated_at': self.updated_at.isoformat(),
                'updated_by': self.updated_by.username
            }
        }

    def __str__(self):
        return str(self.id)

    def save(self, *args, **kwargs):
        if (self.id and not self._state.adding and
                self.details.get('name') != "redacted"):
            old_id = Identity.objects.get(pk=self.id)

            for address_type in ADDRESS_TYPES:
                old_keys = [str(k) for k in
                            old_id.details.get('addresses', {}).get(
                            address_type, {}).keys()]
                new_keys = [str(k) for k in
                            self.details.get('addresses', {}).get(
                            address_type, {}).keys()]

                if (set(old_keys) != set(new_keys)):
                    from .tasks import fire_metric
                    fire_metric.apply_async(kwargs={
                        "metric_name": 'identities.change.{}.sum'.format(
                            address_type),
                        "metric_value": 1.0
                    })

        super(Identity, self).save(*args, **kwargs)

    def remove_details(self, user):
        updated_details = {}
        for attribute, value in self.details.items():
            if attribute != "addresses":
                updated_details[attribute] = "redacted"
            else:
                updated_details[attribute] = {}
        self.details = updated_details
        self.communicate_through = None
        self.updated_by = user
        self.save()

    def optout_address(self, scope, address_type=None, address=None):
        # for each address type (e.g. email, msisdn, etc.)
        for cur_address_type, addresses in self.details["addresses"].items():
            if scope == "all" or cur_address_type == address_type:
                # for each address value (e.g. foo1@bar.com, +27123, etc.)
                for cur_address, cur_details in addresses.items():
                    if scope == "all" or cur_address == address:
                        cur_details["optedout"] = True
        self.save()

    def optin_address(self, address_type=None, address=None):
        self.details["addresses"][address_type][address]["optedout"] = False
        self.save()

    def get_addresses_list(self, address_type, default_only=False):
        response = []
        if "addresses" in self.details:
            if address_type in self.details['addresses']:
                addresses = self.details["addresses"][address_type]
                for address, metadata in addresses.items():
                    if "optedout" in metadata and metadata["optedout"]:
                        break
                    if default_only:
                        # look for default
                        if len(addresses.keys()) > 1:
                            # more than one address, look for default flag
                            if "default" in metadata and metadata["default"]:
                                response.append(address)
                        else:
                            # if only one address its assumed default
                            response.append(address)
                    else:
                        response.append(address)
        return response


@python_2_unicode_compatible
class OptIn(models.Model):
    """An opt-in"""
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False,
                          help_text="UUID of this opt-in request.")
    identity = models.ForeignKey(Identity, null=True,
                                 help_text="UUID for the identity opting in.")
    address_type = models.CharField(
        null=False, max_length=50, default="",
        help_text="Address type used to identify the identity.")
    address = models.CharField(
        null=False, max_length=255, default="",
        help_text="Address used to identify the identity.")
    request_source = models.CharField(
        null=False, max_length=100,
        help_text="Service that the optin was requested from.")
    requestor_source_id = models.CharField(
        null=True, max_length=500,
        help_text="ID for the user requesting the optin on the service that"
        " it was requested from. Ideally a UUID.")
    created_at = models.DateTimeField(auto_now_add=True,
                                      help_text="Time request was received.")
    created_by = models.ForeignKey(User, related_name='optin_created',
                                   null=True,
                                   help_text="User creating the OptIn")

    user = property(lambda self: self.created_by)

    def __str__(self):
        return str(self.id)

    def clean(self):
        """
        Don't allow optins for non existant addresses or ambiguous ones
        This is a duplicate of the view code for DRF to stop future
        internal Django implementations breaking.
        """
        if self.identity is None:
            identities = Identity.objects.filter_by_addr(
                self.address_type, self.address)
            if len(identities) == 0:
                raise ValidationError(
                    "There is no identity with this address.")
            if len(identities) > 1:
                raise ValidationError(
                    "There are multiple identities with this address. "
                    "Please use an explict identity.")


@receiver(pre_save, sender=OptIn)
def optin_saved(sender, instance, **kwargs):
    """
    This is a duplicte of the view code for DRF to stop future
    internal Django implementations breaking.
    """
    if instance.identity is None:
        # look up using the address_type and address
        identities = Identity.objects.filter_by_addr(
            instance.address_type,
            instance.address)
        if identities.count() == 1:
            instance.identity = identities[0]


@receiver(post_save, sender=OptIn)
def handle_optin(sender, instance, created, **kwargs):
    if created is False or instance.identity is None:
        return

    identity = instance.identity

    raw_hook_event.send(
        sender=None,
        event_name='optin.requested',
        payload={
            'identity': str(identity.id),
            'identity_details': identity.details,
            'optin_address_type': instance.address_type,
            'optin_address': instance.address
        },
        user=instance.user,
        send_hook_meta=False
    )

    identity.optin_address(address_type=instance.address_type,
                           address=instance.address)


@python_2_unicode_compatible
class OptOut(models.Model):
    """An optout"""
    OPTOUT_TYPE_CHOICES = (
        ('stop', "No communication on address"),
        ('stopall', "No communication on all addresses"),
        ('unsubscribe', "Unsubcribe"),
        ('forget', "Forget")
    )
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False,
                          help_text="UUID of this optout request.")
    identity = models.ForeignKey(Identity, null=True,
                                 help_text="UUID for the identity opting out.")
    optout_type = models.CharField(
        null=False, max_length=20, default="stop", choices=OPTOUT_TYPE_CHOICES,
        help_text="Type of optout request.")
    address_type = models.CharField(
        null=False, max_length=50, default="",
        help_text="Address type used to identify the identity.")
    address = models.CharField(
        null=False, max_length=255, default="",
        help_text="Address used to identify the identity.")
    request_source = models.CharField(
        null=False, max_length=100,
        help_text="Service that the optout was requested from.")
    requestor_source_id = models.CharField(
        null=True, max_length=500,
        help_text="ID for the user requesting the optout on the service that"
        " it was requested from. Ideally a UUID.")
    reason = models.CharField(
        null=True, max_length=200,
        help_text="Optional reason (e.g. 'not interested')")
    created_at = models.DateTimeField(auto_now_add=True,
                                      help_text="Time request was received.")
    created_by = models.ForeignKey(User, related_name='optout_created',
                                   null=True,
                                   help_text="User creating the OptOut")

    user = property(lambda self: self.created_by)

    def __str__(self):
        return str(self.id)

    def clean(self):
        """
        Don't allow optouts for non existant addresses or ambiguous ones
        This is a duplicate of the view code for DRF to stop future
        internal Django implementations breaking.
        """
        if self.identity is None:
            identities = Identity.objects.filter_by_addr(
                self.address_type, self.address)
            if len(identities) == 0:
                raise ValidationError(
                    "There is no identity with this address.")
            if len(identities) > 1:
                raise ValidationError(
                    "There are multiple identities with this address. "
                    "Please use an explict identity.")


@python_2_unicode_compatible
class DetailKey(models.Model):
    """
    This is a list of all unique keys in the details column of the Identity
    model. Used to help build filters. Populated by post_save triggers.
    """
    key_name = models.CharField(null=False, max_length=200, primary_key=True)
    created_at = models.DateTimeField(auto_now_add=True)

    def __str__(self):
        return str(self.key_name)


@receiver(pre_save, sender=OptOut)
def optout_saved(sender, instance, **kwargs):
    """
    This is a duplicte of the view code for DRF to stop future
    internal Django implementations breaking.
    """
    if instance.identity is None:
        # look up using the address_type and address
        identities = Identity.objects.filter_by_addr(
            instance.address_type,
            instance.address)
        if identities.count() == 1:
            instance.identity = identities[0]


@receiver(post_save, sender=OptOut)
def handle_optout(sender, instance, created, **kwargs):
    if created is False or instance.identity is None:
        return

    identity = instance.identity

    raw_hook_event.send(
        sender=None,
        event_name='optout.requested',
        payload={
            'identity': str(identity.id),
            'identity_details': identity.details,
            'optout_type': instance.optout_type,
            'optout_reason': instance.reason,
            'optout_source': instance.request_source,
        },
        user=instance.user,
        send_hook_meta=False
    )

    if instance.optout_type == "forget":
        identity.remove_details(instance.user)
    elif instance.optout_type == "stop":
        identity.optout_address(scope="single",
                                address_type=instance.address_type,
                                address=instance.address)
    elif instance.optout_type == "stopall":
        identity.optout_address(scope="all")


def get_or_incr_cache(key, func):
    """
    Used to either get a value from the cache, or if the value doesn't exist
    in the cache, run the function to get a value to use to populate the cache
    """
    value = cache.get(key)
    if value is None:
        value = func()
        cache.set(key, value)
    else:
        cache.incr(key)
        value += 1
    return value


@receiver(post_save, sender=OptOut)
def fire_optout_metric(sender, instance, created, **kwargs):
    if created is False or instance.identity is None:
        return

    from .tasks import fire_metric

    fire_metric.apply_async(kwargs={
        "metric_name": 'optout.sum',
        "metric_value": 1.0
    })

    total_key = 'optout.total.last'
    total = get_or_incr_cache(
        total_key,
        OptOut.objects.all().count)
    fire_metric.apply_async(kwargs={
        'metric_name': total_key,
        'metric_value': total,
    })


@receiver(post_save, sender=Identity)
def fire_metrics_if_new(sender, instance, created, **kwargs):
    from .tasks import fire_metric
    if created:
        fire_metric.apply_async(kwargs={
            "metric_name": 'identities.created.sum',
            "metric_value": 1.0
        })


@receiver(post_save, sender=Identity)
def fire_detailkeys_if_new(sender, instance, created, **kwargs):
    from .tasks import populate_detail_key
    if created and instance.details is not None:
        populate_detail_key.apply_async(kwargs={
            "key_names": instance.details.keys()
        })



import json
import uuid
import requests
from celery.task import Task
from django.conf import settings
from go_http.metrics import MetricsApiClient
from .models import Identity, DetailKey
from seed_papertrail.decorators import papertrail


class DeliverHook(Task):
    def run(self, target, payload, instance_id=None, hook_id=None, **kwargs):
        """
        target:     the url to receive the payload.
        payload:    a python primitive data structure
        instance_id:   a possibly None "trigger" instance ID
        hook_id:       the ID of defining Hook object
        """
        requests.post(
            url=target,
            data=json.dumps(payload),
            headers={
                'Content-Type': 'application/json',
                'Authorization': 'Token %s' % settings.HOOK_AUTH_TOKEN
            }
        )


def deliver_hook_wrapper(target, payload, instance, hook):
    if instance is not None:
        if isinstance(instance.id, uuid.UUID):
            instance_id = str(instance.id)
        else:
            instance_id = instance.id
    else:
        instance_id = None
    kwargs = dict(target=target, payload=payload,
                  instance_id=instance_id, hook_id=hook.id)
    DeliverHook.apply_async(kwargs=kwargs)


def get_metric_client(session=None):
    return MetricsApiClient(
        auth_token=settings.METRICS_AUTH_TOKEN,
        api_url=settings.METRICS_URL,
        session=session)


class FireMetric(Task):

    """ Fires a metric using the MetricsApiClient
    """
    name = "seed_identity_store.identities.tasks.fire_metric"

    @papertrail.debug(name)
    def run(self, metric_name, metric_value, session=None, **kwargs):
        metric_value = float(metric_value)
        metric = {
            metric_name: metric_value
        }
        try:
            metric_client = get_metric_client(session=session)
            metric_client.fire(metric)
            return "Fired metric <%s> with value <%s>" % (
                metric_name, metric_value)
        except (requests.exceptions.HTTPError,) as e:
            return "Failed to fire metric <%s> with value <%s> because %s" % (
                metric_name, metric_value, e)


fire_metric = FireMetric()


class ScheduledMetrics(Task):

    """ Fires off tasks for all the metrics that should run
        on a schedule
    """
    name = "seed_identity_store.subscriptions.tasks.scheduled_metrics"

    @papertrail.debug(name)
    def run(self, **kwargs):
        globs = globals()  # execute globals() outside for loop for efficiency
        for metric in settings.METRICS_SCHEDULED_TASKS:
            globs[metric].apply_async()

        return "%d Scheduled metrics launched" % len(
            settings.METRICS_SCHEDULED_TASKS)


scheduled_metrics = ScheduledMetrics()


class FireCreatedLast(Task):

    """ Fires last created subscriptions count
    """
    name = "seed_identity_store.subscriptions.tasks.fire_created_last"

    @papertrail.debug(name)
    def run(self):
        created_identities = Identity.objects.all().count()
        return fire_metric.apply_async(kwargs={
            "metric_name": 'identities.created.last',
            "metric_value": created_identities
        })


fire_created_last = FireCreatedLast()


class PopulateDetailKey(Task):

    """ Fires last created subscriptions count
    """
    name = "seed_identity_store.identities.tasks.populate_detail_key"

    @papertrail.debug(name)
    def run(self, key_names):
        existing_keys = DetailKey.objects.values_list('key_name', flat=True)
        # get key_names NOT in existing_keys
        new_items = [n for n in key_names if n not in existing_keys]
        for i in new_items:
            DetailKey.objects.create(key_name=i)
        return "Added <%s> new DetailKey records" % len(new_items)


populate_detail_key = PopulateDetailKey()



from django.conf import settings


def get_available_metrics():
    available_metrics = []
    available_metrics.extend(settings.METRICS_REALTIME)
    available_metrics.extend(settings.METRICS_SCHEDULED)

    return available_metrics



from rest_framework import viewsets, generics, mixins, status
from rest_framework.exceptions import ValidationError
from rest_framework.permissions import IsAuthenticated, IsAdminUser
from rest_framework.views import APIView
from rest_framework.response import Response
from rest_framework.authtoken.models import Token
from rest_framework import filters
from rest_hooks.models import Hook
from django.contrib.auth.models import User, Group
from django.core.exceptions import FieldError
from .models import Identity, OptOut, OptIn, DetailKey
from .serializers import (UserSerializer, GroupSerializer, AddressSerializer,
                          IdentitySerializer, OptOutSerializer, HookSerializer,
                          CreateUserSerializer, OptInSerializer)
from seed_identity_store.utils import get_available_metrics
from .tasks import scheduled_metrics
import django_filters


class UserViewSet(viewsets.ReadOnlyModelViewSet):
    """ API endpoint that allows users to be viewed or edited.
    """
    permission_classes = (IsAuthenticated,)
    queryset = User.objects.all()
    serializer_class = UserSerializer


class GroupViewSet(viewsets.ReadOnlyModelViewSet):
    """ API endpoint that allows groups to be viewed or edited.
    """
    permission_classes = (IsAuthenticated,)
    queryset = Group.objects.all()
    serializer_class = GroupSerializer


class UserView(APIView):
    """ API endpoint that allows users creation and returns their token.
    Only admin users can do this to avoid permissions escalation.
    """
    permission_classes = (IsAdminUser,)

    def post(self, request):
        '''Create a user and token, given an email. If user exists just
        provide the token.'''
        serializer = CreateUserSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        email = serializer.validated_data.get('email')
        try:
            user = User.objects.get(username=email)
        except User.DoesNotExist:
            user = User.objects.create_user(email, email=email)
        token, created = Token.objects.get_or_create(user=user)

        return Response(
            status=status.HTTP_201_CREATED, data={'token': token.key})


class IdentityFilter(filters.FilterSet):
    """Filter for identities created, using ISO 8601 formatted dates"""
    created_from = django_filters.IsoDateTimeFilter(name="created_at",
                                                    lookup_type="gte")
    created_to = django_filters.IsoDateTimeFilter(name="created_at",
                                                  lookup_type="lte")
    updated_from = django_filters.IsoDateTimeFilter(name="updated_at",
                                                    lookup_type="gte")
    updated_to = django_filters.IsoDateTimeFilter(name="updated_at",
                                                  lookup_type="lte")

    class Meta:
        model = Identity
        fields = ['details', 'communicate_through', 'operator',
                  'created_at', 'created_by', 'updated_at', 'updated_by']


class IdentityViewSet(viewsets.ModelViewSet):
    """ API endpoint that allows identities to be viewed or edited.
    """
    permission_classes = (IsAuthenticated,)
    queryset = Identity.objects.all()
    serializer_class = IdentitySerializer
    filter_class = IdentityFilter

    def perform_create(self, serializer):
        serializer.save(created_by=self.request.user,
                        updated_by=self.request.user)

    def perform_update(self, serializer):
        serializer.save(updated_by=self.request.user)


class IdentitySearchList(generics.ListAPIView):
    permission_classes = (IsAuthenticated,)
    serializer_class = IdentitySerializer

    def get_queryset(self):
        """
        This view should return a list of all the Identities
        for the supplied query parameters. The query parameters
        should be in the form:
        {"address_type": "address"}
        e.g.
        {"msisdn": "+27123"}
        {"email": "foo@bar.com"}

        A special query paramater "include_inactive" can also be passed
        as False to prevent returning identities for which the addresses
        have been set to "inactive"
        e.g.
        {"include_inactive": False}
        """
        query_params = list(self.request.query_params.keys())

        # variable that stores criteria to filter identities by
        filter_criteria = {}
        # variable that stores a list of addresses that should be active
        # if the special filter is passed in
        exclude_if_address_inactive = []

        # Determine from param "include_inactive" whether inactive identities
        # should be included in the search results
        if "include_inactive" in query_params:
            if self.request.query_params["include_inactive"] in [
              "False", 'false', False]:
                include_inactive = False
            else:
                include_inactive = True
        else:
            include_inactive = True  # default to True

        # Compile a list of criteria to filter the identities by, based on the
        # query parameters
        for filter in query_params:
            if filter == "include_inactive":
                # Don't add the special param to the filter_criteria
                pass
            elif filter.startswith("details__addresses__"):
                # Edit the query_param to evaluate the key instead of the value
                # and add it to the filter_criteria
                filter_criteria[filter + "__has_key"] = \
                    self.request.query_params[filter]

                # Add the address to the list of addresses that should not
                # be inactive (tuple e.g ("msisdn", "+27123"))
                if include_inactive is False:
                    exclude_if_address_inactive.append(
                        (filter.replace("details__addresses__", ""),
                         self.request.query_params[filter])
                    )
            else:
                # Add the normal params to the filter criteria
                filter_criteria[filter] = self.request.query_params[filter]

        identities = Identity.objects.filter(**filter_criteria)

        if include_inactive is False:
            # Check through all the identities and exclude ones where the
            # addresses are inactive
            for identity in identities:
                for param in exclude_if_address_inactive:
                    q_key = identity.details["addresses"][param[0]][param[1]]
                    if ('inactive' in q_key and q_key['inactive'] in [True, 'True', 'true']):  # noqa
                        identities = identities.exclude(
                            id=identity.id)

        return identities


class Address(object):
    def __init__(self, address):
        self.address = address


class IdentityAddresses(generics.ListAPIView):
    permission_classes = (IsAuthenticated,)
    serializer_class = AddressSerializer

    def get_queryset(self):
        """
        This view should return a list of all the addresses the identity has
        for the supplied query parameters.
        Currently only supports address_type and default params
        Always excludes addresses with optedout = True
        """
        identity_id = self.kwargs['identity_id']
        address_type = self.kwargs['address_type']
        use_ct = 'use_communicate_through' in self.request.query_params
        default_only = 'default' in self.request.query_params
        if use_ct:
            identity = Identity.objects.select_related('communicate_through')\
                .get(id=identity_id)
            if identity.communicate_through is not None:
                identity = identity.communicate_through
        else:
            identity = Identity.objects.get(id=identity_id)
        addresses = identity.get_addresses_list(address_type, default_only)
        return [Address(addr) for addr in addresses]


class OptInViewSet(mixins.CreateModelMixin, viewsets.GenericViewSet):
    """ API endpoint that allows opt-ins to be created.
    """
    permission_classes = (IsAuthenticated,)
    queryset = OptIn.objects.all()
    serializer_class = OptInSerializer

    def perform_create(self, serializer):
        data = serializer.validated_data
        if "identity" not in data or data["identity"] is None:
            identities = Identity.objects.filter_by_addr(
                data["address_type"], data["address"])
            if len(identities) == 0:
                raise ValidationError(
                    'There is no identity with this address.')
            if len(identities) > 1:
                raise ValidationError(
                    'There are multiple identities with this address.')
            return serializer.save(created_by=self.request.user,
                                   identity=identities[0])
        return serializer.save(created_by=self.request.user)


class OptOutViewSet(mixins.CreateModelMixin, viewsets.GenericViewSet):
    """ API endpoint that allows optouts to be created.
    """
    permission_classes = (IsAuthenticated,)
    queryset = OptOut.objects.all()
    serializer_class = OptOutSerializer

    def perform_create(self, serializer):
        data = serializer.validated_data
        if "identity" not in data or data["identity"] is None:
            identities = Identity.objects.filter_by_addr(
                data["address_type"], data["address"])
            if len(identities) == 0:
                raise ValidationError(
                    'There is no identity with this address.')
            if len(identities) > 1:
                raise ValidationError(
                    'There are multiple identities with this address.')
            return serializer.save(created_by=self.request.user,
                                   identity=identities[0])
        return serializer.save(created_by=self.request.user)


class OptOutSearchList(generics.ListAPIView):
    permission_classes = (IsAuthenticated,)
    serializer_class = OptOutSerializer

    def get_queryset(self):
        try:
            return OptOut.objects.filter(**self.request.query_params.dict())
        except FieldError as e:
            raise ValidationError(str(e))


class HookViewSet(viewsets.ModelViewSet):
    """ Retrieve, create, update or destroy webhooks.
    """
    permission_classes = (IsAuthenticated,)
    queryset = Hook.objects.all()
    serializer_class = HookSerializer

    def perform_create(self, serializer):
        serializer.save(user=self.request.user)


class MetricsView(APIView):

    """ Metrics Interaction
        GET - returns list of all available metrics on the service
        POST - starts up the task that fires all the scheduled metrics
    """
    permission_classes = (IsAuthenticated,)

    def get(self, request, *args, **kwargs):
        status = 200
        resp = {
            "metrics_available": get_available_metrics()
        }
        return Response(resp, status=status)

    def post(self, request, *args, **kwargs):
        status = 201
        scheduled_metrics.apply_async()
        resp = {"scheduled_metrics_initiated": True}
        return Response(resp, status=status)


class HealthcheckView(APIView):

    """ Healthcheck Interaction
        GET - returns service up - getting auth'd requires DB
    """
    permission_classes = (IsAuthenticated,)

    def get(self, request, *args, **kwargs):
        import seed_identity_store
        import django
        import rest_framework
        status = 200
        resp = {
            "up": True,
            "result": {
                "database": "Accessible",
                "version": seed_identity_store.__version__,
                "libraries": {
                    "django": django.__version__,
                    "djangorestframework": rest_framework.__version__
                }
            }
        }
        return Response(resp, status=status)


class DetailKeyView(APIView):

    """ DetailKey retrieval for filter views
        GET - returns list of all available key_names in DetailKey model
    """
    permission_classes = (IsAuthenticated,)

    def get(self, request, *args, **kwargs):
        status = 200
        key_names = DetailKey.objects.values_list('key_name', flat=True)
        resp = {
            "key_names": key_names
        }
        return Response(resp, status=status)



from django.contrib import admin

from .models import Identity, OptOut, OptIn


class IdentityAdmin(admin.ModelAdmin):
    list_display = ["id", "created_at", "updated_at", "created_by",
                    "updated_by"]
    list_filter = ["created_at"]


class OptOutAdmin(admin.ModelAdmin):
    list_display = ["id", "identity", "optout_type", "address_type", "address",
                    "reason", "created_at", "created_by"]
    list_filter = ["optout_type", "address_type", "reason", "created_at",
                   "created_by"]
    search_fields = ["identity__id", "address"]


class OptInAdmin(admin.ModelAdmin):
    list_display = ["id", "identity", "address_type", "address",
                    "created_at", "created_by"]
    list_filter = ["address_type", "created_at", "created_by"]
    search_fields = ["identity__id", "address"]


admin.site.register(Identity, IdentityAdmin)
admin.site.register(OptOut, OptOutAdmin)
admin.site.register(OptIn, OptInAdmin)
